/**
  *  \file planetscentral/host/host.cc
  *  \brief Host server main program
  */

#include <stdexcept>
#include <sys/fcntl.h>
#include <signal.h>
#include <unistd.h>
#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include "planetscentral/host/host.h"
#include "int/args.h"
#include "int/array.h"
#include "planetscentral/dbclient/dbhash.h"
#include "planetscentral/dbclient/dbset.h"
#include "planetscentral/dbclient/dbsubtree.h"
#include "planetscentral/dbclient/reconnect.h"
#include "planetscentral/host/cmdcron.h"
#include "planetscentral/host/cmdgame.h"
#include "planetscentral/host/cmdhost.h"
#include "planetscentral/host/cmdplayer.h"
#include "planetscentral/host/cmdsched.h"
#include "planetscentral/host/cron.h"
#include "planetscentral/host/db.h"
#include "planetscentral/host/runner.h"
#include "planetscentral/host/schedule.h"
#include "planetscentral/osutil.h"
#include "server/config.h"
#include "server/net.h"
#include "util/misc.h"
#include "planetscentral/dbclient/routerclient.h"
#include "planetscentral/dbclient/mailclient.h"

static string_t arg_listen_host = "127.0.0.1";
static string_t arg_listen_port = "7777";
static string_t arg_basedir = ".";
static int32_t arg_threads = 5;

Runner host_runner;
Runner checkturn_runner;
Mutex output_mutex;
DbReadyClient database_connection;
ReconnectingConnection database_rc(database_connection, output_mutex, "127.0.0.1", "6379");
DbClient file_connection;
ReconnectingConnection file_rc(file_connection, output_mutex, "127.0.0.1", "7776");
DbClient userfile_connection;
ReconnectingConnection userfile_rc(userfile_connection, output_mutex, "127.0.0.1", "9998");
MailClient mail_connection;
ReconnectingConnection mail_rc(mail_connection, output_mutex, "127.0.0.1", "21212");
DbClient talk_connection;
ReconnectingConnection talk_rc(talk_connection, output_mutex, "127.0.0.1", "5555");
Cron cron(host_runner, database_connection, database_rc);
RouterClient router_client;

static void
processConfig(string_t key, string_t value)
{
    if (key == "HOST.HOST") {
        /* @q Host.Host:Str (Config)
           Listen address. */
        arg_listen_host = value;
    } else if (key == "HOST.PORT") {
        /* @q Host.Port:Int (Config)
           Port number. */
        arg_listen_port = value;
    } else if (key == "HOST.TIMESCALE") {
        /* @q Host.TimeScale:Int (Config)
           Unix-time-to-{@type Time}-conversion.
           By default, this value is 60, making a time step of 1 equal to a minute.
           For testing, this value can be lowered to make the system run faster
           (i.e. at 1, a daily game runs every 24 minutes, not 24 hours). */
        int n;
        if (stringValue(value, n)) {
            setTimeScale(n);
        }
    } else if (key == "HOSTFILE.BASEDIR") {
        arg_basedir = value;
    } else if (key == "HOST.THREADS") {
        /* @q Host.Threads:Int (Config)
           Number of threads (=maximum number of parallel connections) */
        stringValue(value, arg_threads);
    } else if (key == "REDIS.HOST") {
        database_rc.setHost(value);
    } else if (key == "REDIS.PORT") {
        database_rc.setPort(value);
    } else if (key == "HOSTFILE.HOST") {
        file_rc.setHost(value);
    } else if (key == "HOSTFILE.PORT") {
        file_rc.setPort(value);
    } else if (key == "FILE.HOST") {
        userfile_rc.setHost(value);
    } else if (key == "FILE.PORT") {
        userfile_rc.setPort(value);
    } else if (key == "MAILOUT.HOST") {
        mail_rc.setHost(value);
    } else if (key == "MAILOUT.PORT") {
        mail_rc.setPort(value);
    } else if (key == "TALK.HOST") {
        talk_rc.setHost(value);
    } else if (key == "TALK.PORT") {
        talk_rc.setPort(value);
    } else if (key == "ROUTER.HOST") {
        router_client.setHost(value);
    } else if (key == "ROUTER.PORT") {
        router_client.setPort(value);
    }
}

static bool
isPrintable(const string_t& s)
{
    for (string_t::size_type i = 0; i < s.size(); ++i)
        if (!isalnum(uint8_t(s[i])) && s[i] != '/' && s[i] != '.' && s[i] != '_' && s[i] != '-' && s[i] != ':')
            return false;
    return true;
}

static const char*
getHelp(string_t topic)
{
    string_t s = strUCase(topic);
    if (s == "HOST" || s == "MASTER" || s == "TOOL" || s == "SHIPLIST") {
        return "HOST/MASTER/TOOL/SHIPLIST Commands:\n"
            " HOSTADD id path exe kind\n"
            " HOSTCP oldId newId\n"
            " HOSTDEFAULT id\n"
            " HOSTGET id field\n"
            " HOSTLS\n"
            " HOSTRATING id {{SET n|AUTO} {USE|SHOW}|NONE|GET}\n"
            " HOSTRM id\n"
            " HOSTSET id field value\n"
            "All commands also with prefix MASTER, TOOL, SHIPLIST instead of HOST.\n";
    } else if (s == "CRON") {
        return "CRON Commands:\n"
            " CRONGET gid\n"
            " CRONKICK gid\n"
            " CRONLIST [LIMIT n]\n";
    } else if (s == "GAME") {
        return "GAME Commands:\n"
            " NEWGAME\n"
            " CLONEGAME gid [state]\n"
            " GAMEADDTOOL gid toolid\n"
            " GAMECHECKPERM gid uid\n"
            " GAMEGET gid key\n"
            " GAMEGETCC gid key\n"
            " GAMEGETDIR gid\n"
            " GAMEGETNAME gid\n"
            " GAMEGETOWNER gid\n"
            " GAMEGETSTATE gid\n"
            " GAMEGETTYPE gid\n"
            " GAMEGETVC gid\n"
            " GAMELIST [STATE state] [TYPE type] [VERBOSE] [USER uid]\n"
            " GAMELSTOOLS gid\n"
            " GAMERMTOOL gid toolid\n"
            " GAMESET gid key value\n"
            " GAMESETNAME gid owner\n"
            " GAMESETOWNER gid ownerUid\n"
            " GAMESETSTATE gid state\n"
            " GAMESETTYPE gid type\n"
            " GAMESTAT gid\n"
            " GAMETOTALS\n"
            " GAMEUPDATE gid...\n";
    } else if (s == "PLAYER") {
        return "PLAYER commands:\n"
            " PLAYERADD gid uid\n"
            " PLAYERCHECKFILE gid pid name [DIR dir]\n"
            " PLAYERGETDIR gid pid\n"
            " PLAYERJOIN gid slot uid\n"
            " PLAYERLS gid\n"
            " PLAYERRESIGN gid slot uid\n"
            " PLAYERSETDIR gid pid dir\n"
            " PLAYERSTAT gid slot\n"
            " PLAYERSUBST gid slot uid\n";
    } else if (s == "SCHEDULE") {
        return "SCHEDULE Commands:\n"
            " SCHEDULEADD gid schedulespec\n"
            " SCHEDULEDROP gid\n"
            " SCHEDULELIST gid\n"
            " SCHEDULEMOD gid [schedulespec]\n"
            " SCHEDULESET gid schedulespec\n"
            " SCHEDULESHOW gid [TURNLIMIT n|TIMELIMIT n]\n"
            "schedulespec is:\n"
            " STOP|WEEKLY bits|DAILY interval|ASAP\n"
            " DAYTIME n\n"
            " EARLY|NOEARLY\n"
            " DELAY n\n"
            " LIMIT n\n"
            " UNTILTURN n|UNTILTIME n|FOREVER\n";
    } else {
        return "Commands:\n"
            " PING\n"
            " HELP [<topic>]\n"
            " USER <uid>\n"
            " CRON->\n"
            " GAME->\n"
            " HOST->\n"
            " MASTER->\n"
            " PLAYER->\n"
            " SCHEDULE->\n"
            " SHIPLIST->\n"
            " TOOL->\n"
            " TRN <data> [GAME <gid> [SLOT <slot>]] [MAIL <mail>]\n"
            " TRNMARKTEMP <gid> <slot> <state>\n";
    }
}

/***************************** HostConnection ****************************/

HostConnection::HostConnection()
{ }

void
HostConnection::handleRequest()
{
    // Read request
    std::auto_ptr<IntValue> req(readObject(true));
    IntArray* a = dynamic_cast<IntArray*>(req.get());
    if (!a || a->getData()->getDimension(0) == 0) {
        sendError("400 Syntax error");
        return;
    }
    const IntDataSegment& seg = a->getData()->content;

    // Log request
    {
        MutexGuard g(output_mutex);
        if (user.empty()) {
            printf("[cmd]");
        } else {
            printf("[cmd:%s]", user.c_str());
        }
        for (uint32_t i = 0; i < seg.size(); ++i) {
            IntStringValue* sv = dynamic_cast<IntStringValue*>(seg[i]);
            if (sv != 0) {
                if (sv->getValue().empty()) {
                    printf(" ''");
                } else if (sv->getValue().size() < 100 && isPrintable(sv->getValue())) {
                    printf(" %s", sv->getValue().c_str());
                } else {
                    printf(" ...");
                }
            } else {
                printf(" %s", IntValue::toStringOf(seg[i], false).c_str());
            }
        }
        printf("\n");
    }

    string_t cmd = strUCase(IntValue::toStringOf(seg[0], false));
    IntArgBlock args(seg, 1, seg.size()-1);
    handleCommand(cmd, args);
}

void
HostConnection::handleCommand(const string_t& cmd, IntArgBlock& args)
{
    // Command dispatcher
    if (cmd == "PING") {
        /* @q PING (Host Command)
           Responds with PONG. */
        sendSuccess("PONG");
    } else if (cmd == "HELP") {
        /* @q HELP [page:Str] (Host Command)
           Returns a help page.
           @rettype Str */
        IntStringValue sv(getHelp(toStringResult(args.getNext())));
        sendObject(&sv);
    } else if (cmd == "USER") {
        /* @q USER user:UID (Host Command)
           Set context (caller) for following commands on this connection. */
        user = toStringResult(args.getNext());
        sendSuccess("OK");
        // HOST commands
    } else if (cmd == "HOSTADD") {
        doHostAdd(DbSubtree(database_connection, HOST_PROGRAM_ROOT), *this, args);
    } else if (cmd == "HOSTSET") {
        doHostSet(DbSubtree(database_connection, HOST_PROGRAM_ROOT), *this, args);
    } else if (cmd == "HOSTGET") {
        doHostGet(DbSubtree(database_connection, HOST_PROGRAM_ROOT), *this, args);
    } else if (cmd == "HOSTRM") {
        doHostRemove(DbSubtree(database_connection, HOST_PROGRAM_ROOT), *this, args);
    } else if (cmd == "HOSTLS") {
        doHostList(DbSubtree(database_connection, HOST_PROGRAM_ROOT), *this, args);
    } else if (cmd == "HOSTCP") {
        doHostCopy(DbSubtree(database_connection, HOST_PROGRAM_ROOT), *this, args);
    } else if (cmd == "HOSTDEFAULT") {
        doHostDefault(DbSubtree(database_connection, HOST_PROGRAM_ROOT), *this, args);
    } else if (cmd == "HOSTRATING") {
        doHostRating(DbSubtree(database_connection, HOST_PROGRAM_ROOT), *this, args);
        // MASTER commands
    } else if (cmd == "MASTERADD") {
        doHostAdd(DbSubtree(database_connection, MASTER_PROGRAM_ROOT), *this, args);
    } else if (cmd == "MASTERSET") {
        doHostSet(DbSubtree(database_connection, MASTER_PROGRAM_ROOT), *this, args);
    } else if (cmd == "MASTERGET") {
        doHostGet(DbSubtree(database_connection, MASTER_PROGRAM_ROOT), *this, args);
    } else if (cmd == "MASTERRM") {
        doHostRemove(DbSubtree(database_connection, MASTER_PROGRAM_ROOT), *this, args);
    } else if (cmd == "MASTERLS") {
        doHostList(DbSubtree(database_connection, MASTER_PROGRAM_ROOT), *this, args);
    } else if (cmd == "MASTERCP") {
        doHostCopy(DbSubtree(database_connection, MASTER_PROGRAM_ROOT), *this, args);
    } else if (cmd == "MASTERDEFAULT") {
        doHostDefault(DbSubtree(database_connection, MASTER_PROGRAM_ROOT), *this, args);
    } else if (cmd == "MASTERRATING") {
        doHostRating(DbSubtree(database_connection, MASTER_PROGRAM_ROOT), *this, args);
        // SHIPLIST commands
    } else if (cmd == "SHIPLISTADD") {
        doHostAdd(DbSubtree(database_connection, SHIPLIST_PROGRAM_ROOT), *this, args);
    } else if (cmd == "SHIPLISTSET") {
        doHostSet(DbSubtree(database_connection, SHIPLIST_PROGRAM_ROOT), *this, args);
    } else if (cmd == "SHIPLISTGET") {
        doHostGet(DbSubtree(database_connection, SHIPLIST_PROGRAM_ROOT), *this, args);
    } else if (cmd == "SHIPLISTRM") {
        doHostRemove(DbSubtree(database_connection, SHIPLIST_PROGRAM_ROOT), *this, args);
    } else if (cmd == "SHIPLISTLS") {
        doHostList(DbSubtree(database_connection, SHIPLIST_PROGRAM_ROOT), *this, args);
    } else if (cmd == "SHIPLISTCP") {
        doHostCopy(DbSubtree(database_connection, SHIPLIST_PROGRAM_ROOT), *this, args);
    } else if (cmd == "SHIPLISTDEFAULT") {
        doHostDefault(DbSubtree(database_connection, SHIPLIST_PROGRAM_ROOT), *this, args);
    } else if (cmd == "SHIPLISTRATING") {
        doHostRating(DbSubtree(database_connection, SHIPLIST_PROGRAM_ROOT), *this, args);
        // TOOL commands
    } else if (cmd == "TOOLADD") {
        doHostAdd(DbSubtree(database_connection, TOOL_PROGRAM_ROOT), *this, args);
    } else if (cmd == "TOOLSET") {
        doHostSet(DbSubtree(database_connection, TOOL_PROGRAM_ROOT), *this, args);
    } else if (cmd == "TOOLGET") {
        doHostGet(DbSubtree(database_connection, TOOL_PROGRAM_ROOT), *this, args);
    } else if (cmd == "TOOLRM") {
        doHostRemove(DbSubtree(database_connection, TOOL_PROGRAM_ROOT), *this, args);
    } else if (cmd == "TOOLLS") {
        doHostList(DbSubtree(database_connection, TOOL_PROGRAM_ROOT), *this, args);
    } else if (cmd == "TOOLCP") {
        doHostCopy(DbSubtree(database_connection, TOOL_PROGRAM_ROOT), *this, args);
    } else if (cmd == "TOOLRATING") {
        doHostRating(DbSubtree(database_connection, TOOL_PROGRAM_ROOT), *this, args);
        // NEWGAME
    } else if (cmd == "NEWGAME") {
        doNewGame(*this, args);
    } else if (cmd == "CLONEGAME") {
        doCloneGame(*this, args);
        // GAME commands
    } else if (cmd == "GAMESETTYPE") {
        doSetGameType(*this, args);
    } else if (cmd == "GAMESETSTATE") {
        doSetGameState(*this, args);
    } else if (cmd == "GAMESETOWNER") {
        doSetGameOwner(*this, args);
    } else if (cmd == "GAMESETNAME") {
        doSetGameName(*this, args);
    } else if (cmd == "GAMELIST") {
        doListGames(*this, args);
    } else if (cmd == "GAMESTAT") {
        doStatGame(*this, args);
    } else if (cmd == "GAMESET") {
        doSetGameConfig(*this, args);
    } else if (cmd == "GAMEGET") {
        doGetGameConfig(*this, args);
    } else if (cmd == "GAMEMGET") {
        doGetGameConfigMultiple(*this, args);
    } else if (cmd == "GAMEGETCC") {
        doGetGameCache(*this, args);
    } else if (cmd == "GAMEGETSTATE") {
        doGetGameProperty(*this, args, "state");
    } else if (cmd == "GAMEGETTYPE") {
        doGetGameProperty(*this, args, "type");
    } else if (cmd == "GAMEGETOWNER") {
        doGetGameProperty(*this, args, "owner");
    } else if (cmd == "GAMEGETNAME") {
        doGetGameProperty(*this, args, "name");
    } else if (cmd == "GAMEGETDIR") {
        doGetGameProperty(*this, args, "dir");
    } else if (cmd == "GAMECHECKPERM") {
        doCheckGamePermission(*this, args);
    } else if (cmd == "GAMEADDTOOL") {
        doChangeGameTool(*this, args, true);
    } else if (cmd == "GAMERMTOOL") {
        doChangeGameTool(*this, args, false);
    } else if (cmd == "GAMELSTOOLS") {
        doListGameTools(*this, args);
    } else if (cmd == "GAMETOTALS") {
        doGameTotals(*this, args);
    } else if (cmd == "GAMEGETVC") {
        doGameGetVictoryCondition(*this, args);
    } else if (cmd == "GAMEUPDATE") {
        doGameUpdate(*this, args);
        // SCHEDULE commands
    } else if (cmd == "SCHEDULESET") {
        doScheduleSet(*this, args);
    } else if (cmd == "SCHEDULEADD") {
        doScheduleAdd(*this, args);
    } else if (cmd == "SCHEDULEMOD") {
        doScheduleModify(*this, args);
    } else if (cmd == "SCHEDULELIST") {
        doScheduleList(*this, args);
    } else if (cmd == "SCHEDULEDROP") {
        doScheduleDrop(*this, args);
    } else if (cmd == "SCHEDULESHOW") {
        doScheduleShow(*this, args);
        // PLAYER commands
    } else if (cmd == "PLAYERJOIN") {
        doPlayerJoin(*this, args);
    } else if (cmd == "PLAYERSUBST") {
        doPlayerSubst(*this, args);
    } else if (cmd == "PLAYERRESIGN") {
        doPlayerResign(*this, args);
    } else if (cmd == "PLAYERADD") {
        doPlayerAdd(*this, args);
    } else if (cmd == "PLAYERLS") {
        doPlayerList(*this, args);
    } else if (cmd == "PLAYERSTAT") {
        doPlayerStat(*this, args);
    } else if (cmd == "PLAYERSETDIR") {
        doPlayerSetDir(*this, args);
    } else if (cmd == "PLAYERGETDIR") {
        doPlayerGetDir(*this, args);
    } else if (cmd == "PLAYERCHECKFILE") {
        doPlayerCheckFile(*this, args);
    } else if (cmd == "TRN") {
        doTurnUpload(*this, args);
    } else if (cmd == "TRNMARKTEMP") {
        doTurnMarkTemporary(*this, args);
        // CRON commands
    } else if (cmd == "CRONGET") {
        doCronGet(*this, args);
    } else if (cmd == "CRONLIST") {
        doCronList(*this, args);
    } else if (cmd == "CRONKICK") {
        doCronKick(*this, args);
    } else {
        sendError("400 Unknown command");
    }
}

bool
HostConnection::isAdmin() const
{
    return user.empty();
}

void
HostConnection::checkAdmin()
{
    if (!isAdmin()) {
        throw std::runtime_error("403 Permission denied");
    }
}

void
HostConnection::checkPermission(Game& g, Game::PermissionLevel level)
{
    if (!g.hasPermission(user, level)) {
        throw std::runtime_error("403 Permission denied");
    }
}

string_t
HostConnection::getUser() const
{
    return user;
}



static int server_socket;

static void*
serverLoop(void*)
{
    while (1) {
        HostConnection conn;
        if (!conn.accept(server_socket)) {
            perror("accept");
            sleep(10);
        } else {
            try {
                while (!conn.isAtEnd()) {
                    try {
                        conn.handleRequest();
                    }
                    catch (std::exception& e) {
                        {
                            MutexGuard g(output_mutex);
                            printf("[cmd] [exception] %s\n", e.what());
                        }
                        conn.sendError(e.what());
                    }
                    conn.flush();
                }
            }
            catch (std::exception& e) {
                printf("c2host: %s\n", e.what());
            }
            conn.close();
        }
    }
}

int main(int /*argc*/, char** argv)
{
    /* Args */
    bool opt_nocron = false;
    while (const char* p = *++argv) {
        if (p[0] == '-' && p[1] == '-') {
            ++p;
        }
        if (strcmp(p, "-nocron") == 0) {
            opt_nocron = true;
        } else {
            printf("c2host: invalid parameter.\n\n"
                   "Usage: c2host [-nocron]\n");
            return 1;
        }
    }

    /* Unbuffer stdout */
    setvbuf(stdout, 0, _IOLBF, 0);

    /* Configuration */
    loadConfiguration(processConfig);

    /* Change to base directory */
    if (chdir(arg_basedir.c_str()) != 0) {
        perror(arg_basedir.c_str());
        return 1;
    }

    /* Start runners */
    host_runner.start();
    checkturn_runner.start();

    /* Connect to database */
    try {
        database_rc.connectRetry();
        file_rc.connectRetry();
        userfile_rc.connectRetry();
        mail_rc.connectRetry();
        talk_rc.connectRetry();
    }
    catch (std::exception& e) {
        fprintf(stderr, "unable to connect to servers: %s\n", e.what());
        return 1;
    }

    /* Start cron */
    if (opt_nocron) {
        printf("[main] cron will not be run!\n");
    } else {
        cron.start();
    }

    /* Listen */
    server_socket = Bind().withAddress(arg_listen_host, arg_listen_port);
    if (server_socket < 0) {
        fprintf(stderr, "[main] unable to bind socket\n");
        return 1;
    }

    /* OK, main loop */
    printf("[main] listening...\n");
    signal(SIGPIPE, SIG_IGN);
    fcntl(server_socket, F_SETFD, FD_CLOEXEC);
    listen(server_socket, 10);

    /* Create some threads. Create one less than configured; the main
       thread will also serve requests. */
    for (int32_t i = 1; i < arg_threads; ++i) {
        pthread_t handle;
        if (pthread_create(&handle, 0, serverLoop, 0) < 0) {
            perror("[main] unable to create thread");
            return 1;
        }
    }
    serverLoop(0);
    close(server_socket);
}
