/**
  *  \file planetscentral/host/install.cc
  *  \brief Install Game Data
  */

#include <set>
#include <memory>
#include <stdexcept>
#include <cstring>
#include <cstdio>
#include "planetscentral/host/install.h"
#include "cpluslib/format.h"
#include "game/struct.h"
#include "int/args.h"
#include "int/array.h"
#include "io/dirs.h"
#include "io/streams.h"
#include "planetscentral/dbclient/fileclient.h"
#include "planetscentral/host/game.h"
#include "planetscentral/host/host.h"
#include "planetscentral/osutil.h"
#include "planetscentral/dbclient/routerclient.h"
#include "planetscentral/dbclient/reconnect.h"

/** Check file name match. In perl terms, checks $name =~ /$pre[0-9]+$post/. */
static bool
match(const string_t& name, const char* pre, const char* post)
{
    // Check length
    std::size_t preLen = std::strlen(pre);
    std::size_t postLen = std::strlen(post);
    if (name.size() <= preLen + postLen) {
        return false;
    }

    // Check content
    if (name.compare(0, preLen, pre, preLen) != 0 || name.compare(name.size()-postLen, postLen, post, postLen) != 0) {
        return false;
    }
    if (name.find_first_not_of("0123456789", preLen) < name.size() - postLen) {
        return false;
    }
    return true;
}

/** Check for file to be kept by game directory installation. */
bool
isPreciousFile(const string_t& name)
{
    // The idea is that specification files which are recreated by a new
    // install are erased, and all specification files which are not part
    // of the new install are erased. That is, there's no need to keep a
    // hconfig.hst file because if the game contains one, it will be
    // recreated, and if the game does not contain one, there shouldn't
    // be any. So all we need to take care of is database files. Check
    // for the important files known to be created by PCC, Winplan, VPA.
    //
    // This uses the paradigm to delete everything I don't know. It is
    // probably debatable whether that should be the other way around,
    // so we only delete foreign playerX.rsts and keep other files intact.
    return name == "fizz.bin"
        || name == "stat.cc"
        || name == "score.cc"
        || name == "config.cc"
        || name == "config2.cc"
        || match(name, "auto", ".dat")
        || match(name, "chart", ".cc")
        || match(name, "fleet", ".cc")
        || match(name, "notes", ".dat")
        || match(name, "script", ".cc")
        || match(name, "team", ".cc")
        || match(name, "vm", ".cc")
        || match(name, "vpa", ".db");
}

/** List files. Generates a list of files in a directory.
    \param file Filer to check
    \param dirName Directory on filer
    \param files Set of file names (excluding paths) will be produced here */
static void
listFiles(FileClient& file, string_t dirName, std::set<string_t>& files)
{
    // Fetch content.
    std::auto_ptr<IntValue> content(file.getDirectoryContent(dirName));
    IntArray* a = dynamic_cast<IntArray*>(content.get());
    if (a == 0) {
        return;
    }

    // Content is list of string (name) and array (properties).
    // The properties have one "type" property. We only want files.
    IntArgBlock aa(a->getData()->content, 0, a->getData()->getDimension(0));
    while (aa.getNumArgs() > 1) {
        string_t name = toStringResult(aa.getNext());
        IntArray* props = dynamic_cast<IntArray*>(aa.getNext());
        if (props == 0) {
            // Protocol error
        } else {
            // Check properties
            bool isFile = false;
            IntArgBlock pp(props->getData()->content, 0, props->getData()->getDimension(0));
            while (pp.getNumArgs() > 1 && !isFile) {
                string_t propName = toStringResult(pp.getNext());
                string_t propVal  = toStringResult(pp.getNext());
                if (propName == "type" && propVal == "file") {
                    isFile = true;
                }
            }
            if (isFile) {
                files.insert(name);
            }
        }
    }
}

/** Install file.
    \param s File to copy (freshly opened)
    \param file Copy to this filer...
    \param dirName ...into this directory...
    \param name ...under this name */
static void
installFile(Stream& s,
            FileClient& file, string_t dirName,
            string_t name)
{
    int32_t size = s.getSize();
    if (size < 0 || size > 100*1024*1024) {
        // Make sure we don't accidentally crash. This normally does not
        // happen because files come from a trusted source (sort of) which
        // does not create files that big. 100meg still is a lot because
        // it's going to be copied around a little...
        throw std::runtime_error("File size error");
    }
    string_t buffer;
    buffer.resize(size);
    s.readT(&buffer[0], size);
    file.putFile(dirName + "/" + name, buffer);
}

/** Install files from a directory.
    \param srcDirName Copy all files from this directory...
    \param file ...to this filer...
    \param dirName ...into this directory.
    \param filesToDelete Copied files are removed from here.
    \param except This file is not copied. */
static void
installFiles(string_t srcDirName,
             FileClient& file, string_t dirName,
             std::set<string_t>& filesToDelete,
             string_t except)
{
    // Log message
    {
        MutexGuard g(output_mutex);
        std::printf("[install] copying files from 'host:%s' to 'user:%s'\n", srcDirName.c_str(), dirName.c_str());
    }

    // Do it
    Ptr<Directory> dir = Directory::openDir(srcDirName);
    Ptr<Enumeration<Ptr<DirectoryEntry> > > entries = dir->getEntries();
    while (entries->hasMoreElements()) {
        // A file is accepted if
        // ...it does not start with a '.' (e.g. ".." or ".c2file")
        // ...it is not explicitly excempted
        // ...its name is entirely lowercase, otherwise it won't be accessible to the user
        // ...it actually is a file
        Ptr<DirectoryEntry> e = entries->getNextElement();
        string_t name = e->getBaseName();
        if (!name.empty()
            && name[0] != '.'
            && name != except
            && strLCase(name) == name
            && e->getFileType() == DirectoryEntry::FT_FILE)
        {
            // Must copy this file
            Ptr<Stream> s = e->openFile(Stream::C_READ);
            installFile(*s, file, dirName, name);

            // Must not erase this file
            filesToDelete.erase(name);
        }
    }
}

/** Close PCC2 Web sessions in a directory. Failure is nonfatal:
    If c2router is not running, there's no session that needs to be terminated. */
static void
closeClientSession(const string_t& dirName)
{
    router_client.tryCall(format("CLOSE -WDIR=%s\n", dirName));
}

/** Install game data.
    \param game    Game to install from (locked)
    \param players Whose player data to use
    \param file    File client (locked and set to appropriate user permissions)
    \param dirName Directory name */
void
installGameData(Game& game, GPlayerSet players, FileClient& file, string_t dirName)
{
    // Kill all running PCC2 Web sessions in this directory.
    closeClientSession(dirName);

    // Start by configuring the game properties
    string_t gameState = game.getState();
    file.createDirectoryTree(dirName);
    file.setProperty(dirName, "game", game.getId());
    file.setProperty(dirName, "name", game.getName());
    file.setProperty(dirName, "nofilewarning", 1);
    file.setProperty(dirName, "finished", gameState == "finished" || gameState == "deleted");
    file.setProperty(dirName, "hosttime", 0);    // FIXME: would be nice to have this correct

    // List files
    std::set<string_t> filesToDelete;
    listFiles(file, dirName, filesToDelete);

    // Install files
    string_t gameDir = game.getDirectory();
    installFiles(gameDir + "/out/all", file, dirName, filesToDelete, "playerfiles.zip");
    for (int i = 1; i <= NUM_PLAYERS; ++i) {
        if (players.contains(i)) {
            // Outgoing files
            installFiles(gameDir + "/out/" + itoa(i), file, dirName, filesToDelete, format("player%d.zip", i));

            // Turn files
            string_t trnName = format("player%d.trn", i);
            if (Ptr<Stream> trn = Stream::openFileNT(gameDir + "/in/" + trnName, Stream::C_READ)) {
                installFile(*trn, file, dirName, trnName);
                filesToDelete.erase(trnName);
            }
        }
    }

    // Remove surplus files
    for (std::set<string_t>::iterator i = filesToDelete.begin(); i != filesToDelete.end(); ++i) {
        if (!isPreciousFile(*i)) {
            file.removeFile(dirName + "/" + *i);
        }
    }
}

/** Install a file into multiple players' game directories.
    \param game    Game, locked
    \param players List of players
    \param file    User file client, locked
    \param fileName Name of file
    \param fileContent Content of file */
void
installFileMulti(Game& game, StringList_t& players, FileClient& file, string_t fileName, string_t fileContent, int32_t slot)
{
    int did = 0, fail = 0;
    for (uint32_t i = 0; i < players.size(); ++i) {
        // Try each player in turn with his own permissions.
        // Don't fail if the player has messed with the directory.
        string_t dirName = game.getPlayerConfig(players[i], "gameDir");
        if (!dirName.empty()) {
            file.setUserPermissions(players[i]);
            try {
                if (file.getIntProperty(dirName, "game") == game.getId()) {
                    if (slot != 0) {
                        router_client.tryCall(format("CLOSE -WDIRPL=%s/%d\n", dirName, slot));
                    }
                    file.putFile(dirName + "/" + fileName, fileContent);
                }
                ++did;
            }
            catch (...) {
                ++fail;
            }
        }
    }
    {
        MutexGuard g(output_mutex);
        std::printf("[install] '%s' copied to %d directories, failed %d'\n", fileName.c_str(), did, fail);
    }
}

/** Install or uninstall game files after an (un)subscription.
    \param game   Game, locked
    \param player Changed player
    \param slot   Changed slot
    \param added  true if player was added, false if he was removed */
void
installChangedGameFiles(Game& game, string_t player, int32_t /*slot*/, bool /*added*/)
{
    // Is there a directory?
    string_t dirName = game.getPlayerConfig(player, "gameDir");
    if (dirName.empty()) {
        return;
    }

    // Figure out all subscriptions of that player
    GPlayerSet set = game.getSlotsByPlayer(player);

    // Do it
    ReconnectingGuard g(userfile_rc);
    FileClient file(userfile_connection);
    file.setUserPermissions(player);
    if (set.empty()) {
        // Last unsubscription: unconfigure game
        game.setPlayerConfig(player, "gameDir", string_t());
        uninstallGameData(file, dirName);
        try {
        }
        catch (...) { }
    } else {
        // Some subscriptions remain: regenerate the game directory.
        // We could optimize here, but let's start simple.
        // FIXME: actually we should optimize here. This kills a user's
        // PCC2 Web session - and his turn file! - when someone
        // appoints him as a replacement.
        installGameData(game, set, file, dirName);
    }
}

/** Uninstall game data. Drops the link from the filer to the host
    (but keeps the files there for the user).
    \param file File client, locked and configured for user
    \param dirName directory name to uninstall */
void
uninstallGameData(FileClient& file, string_t dirName)
{
    try {
        // We're leaving an old game directory. Change its properties.
        // However, if anything goes wrong in the process, don't fail.
        // Actually, the only thing we have to do is to retract the
        // 'game' property. But retract the others as well.
        file.setProperty(dirName, "game", 0);
        file.setProperty(dirName, "finished", 0);
        file.setProperty(dirName, "hosttime", 0);
    }
    catch (...) { }
}
