/**
  *  \file planetscentral/talk/topic.cc
  */

#include "planetscentral/talk/topic.h"
#include "planetscentral/talk/db.h"
#include "planetscentral/talk/forum.h"
#include "int/array.h"
#include "cpluslib/misc.h"
#include "planetscentral/talk/message.h"
#include "planetscentral/talk/user.h"

Topic::Topic(DbClient& client, int32_t topicId)
    : topic(DbSubtree(client, THREAD_ROOT).subtree(topicId)),
      topicId(topicId)
{ }

DbHash
Topic::header()
{
    return topic.hashKey("header");
}

DbHashStringField
Topic::subject()
{
    return header().stringField("subject");
}

DbHashIntField
Topic::forumId()
{
    return header().intField("forum");
}

DbHashIntField
Topic::firstPostingId()
{
    return header().intField("firstpost");
}

DbHashStringField
Topic::readPermissions()
{
    return header().stringField("readperm");
}

DbHashStringField
Topic::answerPermissions()
{
    return header().stringField("answerperm");
}

DbHashIntField
Topic::lastPost()
{
    return header().intField("lastpost");
}

DbHashIntField
Topic::lastTime()
{
    return header().intField("lasttime");
}

DbIntSet
Topic::messages()
{
    return topic.intSetKey("messages");
}

DbStringSet
Topic::watchers()
{
    return topic.stringSetKey("watchers");
}

Forum
Topic::forum()
{
    return Forum(topic.getClient(), forumId().get());
}

/** Remove this topic. Removes all messages. */
void
Topic::remove()
{
    // This is a simple, not too efficient implementation: just remove all messages.
    // The last Message::remove() will call Topic::removeEmpty().
    // This contains a few more database lookups than necessary (e.g. the mapping from
    // the message to the topic), but then, topic removal hopefully isn't frequent.
    IntList_t msg;
    messages().getAll(msg);
    for (uint32_t i = 0; i < msg.size(); ++i) {
        Message(topic.getClient(), msg[i]).remove();
    }
}

/** Remove this empty topic.
    Must be called for empty topics only. */
void
Topic::removeEmpty()
{
    Forum f(forum());

    // Remove from list
    f.topics().remove(topicId);
    f.stickyTopics().remove(topicId);

    // Remove watchers
    StringList_t w;
    watchers().getAll(w);
    for (uint32_t i = 0; i < w.size(); ++i) {
        User u(topic.getClient(), w[i]);
        u.watchedTopics().remove(topicId);
        u.notifiedTopics().remove(topicId);
    }

    // Remove topic
    header().remove();
    messages().remove();
    watchers().remove();
}

bool
Topic::isSticky()
{
    return header().intField("sticky").get();
}

void
Topic::setSticky(bool enable)
{
    bool isSticky = header().getIntField("sticky");
    if (isSticky != enable) {
        header().setField("sticky", enable);
        Forum f(topic.getClient(), forumId().get());
        if (enable) {
            // normal -> sticky
            f.topics().moveTo(topicId, f.stickyTopics());
        } else {
            // sticky -> normal
            f.stickyTopics().moveTo(topicId, f.topics());
        }
    }
}

int32_t
Topic::getId() const
{
    return topicId;
}

int32_t
Topic::allocateTopic(DbClient& client)
{
    return ++DbSubtree(client, THREAD_ROOT).intKey("id");
}

bool
Topic::exists()
{
    return header().exists();
}

IntValue*
Topic::describe()
{
    /* @type TalkThreadInfo
       Information about a forum thread (topic).
       This is an excerpt of {thread:$TID:header}.

       @key subject:Str (subject)
       @key forum:FID (forum Id)
       @key firstpost:MID (MID of first posting)
       @key lastpost:MID (MID of last posting)
       @key lasttime:Time (time of last posting)
       @key sticky:Int (1 if thread is sticky) */

    // FIXME: can we use HMGET?
    const char*const fields[] = { "subject", "forum", "firstpost", "lastpost", "lasttime", "sticky" };
    Ptr<IntArrayData> ad = new IntArrayData();
    for (std::size_t i = 0; i < countof(fields); ++i) {
        ad->content.pushNew(makeStringValue(fields[i]));
        ad->content.pushNew(header().getFieldRaw(fields[i]));
    }
    ad->addDimension(ad->content.size());
    return new IntArray(ad);
}
